<?php

namespace QUI\NodeDev;

use QUI;
use QUI\Exception;
use League\CLImate\CLImate;

use function array_merge;
use function chdir;
use function file_exists;
use function file_get_contents;
use function file_put_contents;
use function implode;
use function is_string;
use function system;
use function version_compare;

class Console extends QUI\System\Console\Tool
{
    protected bool $systemTool = true;

    public function __construct()
    {
        $this->setName('nodejs')
            ->setDescription('NodeJS project helper');
    }

    /**
     * Execute the NodeJS project helper tool.
     *
     * This method runs the NodeJS project helper tool, handling different commands
     * and options. It creates the node folder and packages.json file, and then
     * executes the specified command.
     *
     * @return void
     * @throws \Exception if an error occurs during execution
     */
    public function execute(): void
    {
        try {
            $this->getNodeJsCommand();
            $this->getNPMCommand();
        } catch (\Exception $Exception) {
            $this->writeLn($Exception->getMessage(), 'red');
            $this->writeLn();
            $this->resetColor();

            return;
        }

        $args = array_slice($_SERVER['argv'], 1);

        // create node folder
        $nodeFolder = self::getNodeProjectPath();
        QUI\Utils\System\File::mkdir($nodeFolder);

        // create packages json
        $this->createPackagesJson();

        if (!isset($args[0])) {
            $climate = new CLImate();

            $climate->br();
            $climate->blue()->bold()->out('NodeJS project path:');
            $climate->out($nodeFolder);

            $climate->br();
            $climate->blue()->bold()->out('NodeJS examples:');
            $climate->br();

            $climate->table([
                ['nodejs node [command]', 'Run a command using nodejs directly'],
                ['nodejs node -v', 'Display the Node.js version']
            ]);

            $climate->br();
            $climate->blue()->bold()->out('NPM examples:');
            $climate->br();

            $climate->table([
                ['nodejs npm [command]', 'Run a command using npm'],
                ['nodejs npm init', 'Initialize a new Node.js project'],
                ['nodejs npm install', 'Install or update Node.js packages'],
                ['nodejs npm install [package]', 'Install a specific package'],
                ['nodejs npm run build', 'Build the Node.js project'],
                ['nodejs npm start', 'Start the Node.js application']
            ]);

            $climate->br();
            $climate->blue()->bold()->out('NPX examples:');
            $climate->br();

            $climate->table([
                ['nodejs npx [command]', 'Run a command using npx'],
                ['nodejs npx tailwindcss init', 'Initialize a new Tailwind CSS project']
            ]);
            return;
        }

        switch ($args[0]) {
            case 'node':
                $args[0] = $this->getNodeJsCommand();
                break;

            case 'npm':
                $args[0] = $this->getNPMCommand();
                break;

            case 'npx':
                $args[0] = $this->getNPXCommand();
                break;
        }

        $command = implode(' ', $args);


        $this->writeLn('change folder to ' . $nodeFolder);
        chdir($nodeFolder);

        $this->writeLn('Execute command: ' . $command);
        $this->writeLn();
        $this->writeLn();
        system($command);
    }

    /**
     * Returns the nodejs command based on the configuration.
     *
     * @return string
     * @throws QUI\Exception if nodejs could not be found
     */
    public function getNodeJsCommand(): string
    {
        $Config = QUI::getPackage('quiqqer/node-development')->getConfig();
        $nodejs = $Config?->getValue('commands', 'nodejs');

        if (!is_string($nodejs)) {
            $nodejs = '';
        }

        if (empty($nodejs)) {
            $nodejs = 'nodejs';

            if (!QUI\Utils\System::isSystemFunctionCallable($nodejs)) {
                $nodejs = '';
            }
        }

        if (empty($nodejs)) {
            $nodejs = 'node';

            if (!QUI\Utils\System::isSystemFunctionCallable($nodejs)) {
                $nodejs = '';
            }
        }

        if (empty($nodejs)) {
            throw new QUI\Exception(
                'nodejs could not be found. please install nodejs or setup the binary path at the config'
            );
        }

        return $nodejs;
    }


    /**
     * Returns the npm command based on the configuration.
     *
     * @return string
     * @throws QUI\Exception if npm could not be found
     */
    public static function getNPMCommand(): string
    {
        $Config = QUI::getPackage('quiqqer/node-development')->getConfig();
        $npm = $Config?->getValue('commands', 'npm');

        if (!is_string($npm)) {
            $npm = '';
        }

        if (empty($npm)) {
            $npm = 'npm';

            if (!QUI\Utils\System::isSystemFunctionCallable($npm)) {
                $npm = '';
            }
        }

        if (empty($npm)) {
            throw new QUI\Exception(
                'npm could not be found. please install npm or setup the binary path at the config'
            );
        }

        return $npm;
    }

    /**
     * Returns the npx command based on the configuration.
     *
     * @return string
     * @throws QUI\Exception if npx could not be found
     */
    public static function getNPXCommand(): string
    {
        $Config = QUI::getPackage('quiqqer/node-development')->getConfig();
        $npx = $Config?->getValue('commands', 'npx');

        if (!is_string($npx)) {
            $npx = '';
        }

        if (empty($npx)) {
            $npx = 'npx';

            if (!QUI\Utils\System::isSystemFunctionCallable($npx)) {
                $npx = '';
            }
        }

        if (empty($npx)) {
            throw new QUI\Exception(
                'npx could not be found. please install npx or setup the binary path at the config'
            );
        }

        return $npx;
    }

    /**
     * Returns the path to the node project directory.
     *
     * @return string
     * @throws Exception
     */
    public static function getNodeProjectPath(): string
    {
        $Package = QUI::getPackage('quiqqer/node-development');
        $varDir = $Package->getVarDir();

        return $varDir . 'bin/nodejs/';
    }

    /**
     * Creates a packages.json file containing the dependencies of the installed packages.
     *
     * This method iterates over the installed packages, reads their package.json files,
     * and merges their dependencies into a single packages.json file.
     *
     * @return void
     * @throws QUI\Exception if the package manager is not available
     */
    protected function createPackagesJson(): void
    {
        $packages = QUI::getPackageManager()->getInstalled();
        $packagesDir = OPT_DIR;
        $require = [];

        // fill require with main packages json dependencies
        $projectPath = self::getNodeProjectPath();

        if (!file_exists($projectPath . 'package.json')) {
            file_put_contents($projectPath . 'package.json', '{}');
        }

        $projectJson = file_get_contents($projectPath . 'package.json');

        if (!is_string($projectJson)) {
            $projectJson = '{}';
        }

        $projectJson = json_decode($projectJson, true) ?? [];

        if (isset($projectJson['dependencies'])) {
            $require = array_merge($projectJson['dependencies'], $require);
        }

        // fill require with package.json dependencies from plugins
        foreach ($packages as $package) {
            $packageDir = $packagesDir . $package['name'];
            $packageJson = $packageDir . '/package.json';

            if (!file_exists($packageJson)) {
                continue;
            }

            // get dependencies
            $data = file_get_contents($packageJson);

            if (!is_string($data)) {
                $data = '{}';
            }

            $data = json_decode($data, true) ?? [];

            if (!isset($data['dependencies'])) {
                continue;
            }

            $dependencies = $data['dependencies'];

            foreach ($dependencies as $packageName => $version) {
                if (empty($version)) {
                    $version = '*';
                }

                if (!isset($require[$packageName])) {
                    $require[$packageName] = $version;
                } elseif (version_compare($require[$packageName], $version) == -1) {
                    $require[$packageName] = $version;
                }
            }
        }

        $projectJson['dependencies'] = $require;

        file_put_contents($projectPath . 'package.json', json_encode($projectJson, JSON_PRETTY_PRINT));
    }
}
